"""
	Description:
					Public functions to easy several tasks with Files
	Provides:
					import AspireTUI.UtilsFile as UF
					ret_bool, ret_str = cu.reg_value_get(file,key,var)
					
	========================================================
	Created on:		2024 Jan. 01
	Created by:		Simon Arjuna Erat
	License:		MIT
	URL:			https://www.github.com/sri-arjuna/ASPIRE

	Based on my TUI & SWARM for the BASH shell © 2011
"""
#
#	Essential imports
#
import os as _os
#import sys as _sys
#import re as _re
#import string as _string
from AspireTUI import _MSG
#from . import Classes as _Classes
#from . import strings as _stew
from AspireTUI import tui as _tui
#from typing import Union as _Union
################################################################################################################
#####                                            Basic Checks                                              #####
################################################################################################################
# FileDescriptorOrPath
def exists(filename: str, bVerbose=False, bDual=False):
	"""
	Returns True if filename exists, or both, bool and message.
	
	Examples:
		if _Files.file_exists(filename):
			# Do stuff
			pass
		
		ret_bool, ret_msg = _Files.file_exists(filename, bDual=True)

	"""
	ret_bool = None
	ret_bool = _os.path.exists(filename)
	fn = _os.path.basename(filename)
	ret_msg = f"{_MSG.word_found} {_MSG.word_filesystem_file}: {fn}"

	if ret_bool:
		ret_msg = _MSG.file_exists
	else:
		ret_msg = _MSG.file_not_found
	# Display things, if desired
	if bVerbose:
		_tui.status(ret_bool, ret_msg)
	# Return 
	if bDual:
		return ret_bool, ret_msg
	else:
		return ret_bool

def isfile(filename: str, bVerbose=False, bDual=False):
	"""
	Returns True if it is a file.
	"""
	# Init vars
	ret_bool = False
	fn = _os.path.basename(filename)
	ret_msg = f"{_MSG.word_found} {_MSG.word_filesystem_file}: {fn}"
	# Start checking
	if exists(filename, bVerbose=bVerbose):
		if _os.path.isfile(filename):
			# Its a file
			ret_bool = True
	else:
		raise TypeError(filename)
	# Display things, if desired
	if bVerbose:
		_tui.status(ret_bool, ret_msg, fn)
	# Return 
	if bDual:
		return ret_bool, ret_msg
	else:
		return ret_bool

def isdir(filename: str, bVerbose=False, bDual=False):
	"""
	Returns True if it is a direcctory.
	"""
	# Init vars
	ret_bool = False
	fn = _os.path.basename(filename)
	ret_msg = f"{_MSG.word_found} {_MSG.word_filesystem_dir}: {fn}"
	# Start checking
	if exists(filename, bVerbose=bVerbose):
		if _os.path.isdir(filename):
			# Its a dir
			ret_bool = True
	else:
		raise TypeError(filename)
	# is Verbose?
	if bVerbose:
		_tui.status(ret_bool, ret_msg)
	# Return
	if bDual:
		return ret_bool, ret_msg
	else:
		return ret_bool

def islink(filename: str, bVerbose=False, bDual=False):
	"""
	Returns True if it is a link.
	"""
	# Init vars
	ret_bool = False
	fn = _os.path.basename(filename)
	ret_msg = f"{_MSG.word_found} {_MSG.word_filesystem_link}: {fn}"
	# Start checking
	if exists(filename, bVerbose=bVerbose):
		if _os.path.islink(filename):
			# Its a dir
			ret_bool = True
	else:
		raise TypeError(filename)
	# is Verbose?
	if bVerbose:
		_tui.status(ret_bool, ret_msg)
	# Return
	if bDual:
		return ret_bool, ret_msg
	else:
		return ret_bool

def ismount(filename: str, bVerbose=False, bDual=False):
	"""
	Returns True if it is a direcctory.
	"""
	# Init vars
	ret_bool = False
	fn = _os.path.basename(filename)
	ret_msg = f"{_MSG.word_found} {_MSG.word_filesystem_mount}: {fn}"
	# Start checking
	if exists(filename, bVerbose=bVerbose):
		if _os.path.ismount(filename):
			# Its a dir
			ret_bool = True
	else:
		raise TypeError(filename)
	# is Verbose?
	if bVerbose:
		_tui.status(ret_bool, ret_msg)
	# Return
	if bDual:
		return ret_bool, ret_msg
	else:
		return ret_bool

def is_file_in_use(filename: str):
	"""
	Returns True if file is in use.
	"""
	try:
		with open(filename, 'r') as file:
			pass
		# File is not in use
		return False
	except PermissionError:	# OSError in Python 3.3 and later
		# File is in use
		return True



"""
	Description:
					Internal functions for tui.*, handle string splitting and theme.
	
	Provides:
					put.border(style=header|title)			\n
					put.text(*args, style=header|title)

	========================================================
	Created on:		2023 Nov. 09
	Created by:		Simon Arjuna Erat
	License:		MIT
	URL:			https://www.github.com/sri-arjuna/ASPIRE

	Based on my TUI & SWARM for the BASH shell © 2011
"""
################################################################################################################
#####                                            Shared Imports                                            #####
################################################################################################################
#
#	Essential imports
#
from AspireTUI import _MSG
from AspireTUI import IS_WINDOWS as _IS_WINDOWS
#import os as _os
#import sys as _sys
#import re as _re
import string as _string
#
#	Cross Platform & Advanced usage
#
#import platform as _platform
import msvcrt as _msvcrt
#import shutil as _shutil
import subprocess as _subprocess
#
#	Prepare data structures
#
#from collections import namedtuple as _namedtuple
#from dataclasses import dataclass as _dataclass
from enum import Enum as _Enum
#from enum import EnumMeta as _EnumMeta
#from typing import Union as _Union

################################################################################################################
#####                                            Error Handler                                             #####
################################################################################################################
class _AspireErrorEnum(_Enum):
	THEME_NONE 			= _MSG.theme_none,
	THEME_CANT_READ 	= _MSG.theme_cant_read,
	THEME_EMPTY_VAR		= _MSG.theme_empty_var,
	COLOR_CANT_READ		= _MSG.theme_color_invalid,
	PRINT_COUNT			= _MSG.args_max3,
	STATUS_ID			= _MSG.args_status_first,
	STATUS_COUNT		= _MSG.args_2_status
class _AspireErrorSeverityEnum(_Enum):
	FATAL = _MSG.word_fatal,
	ERROR = _MSG.word_error

def ErrMsg(id: _AspireErrorEnum, severity: _AspireErrorSeverityEnum) -> str:
	print(severity + "\n" + id)
################################################################################################################
#####                                           String Utils (stew)                                        #####
################################################################################################################
def split_needed(txt_len: int, LineLength:int, percentage: int, style="print"):
	if txt_len >= LineLength:
		# Text is too long
		return True
	else:
		# Basicly, no split required, but there are exceptions
		# Specificly, if it is title, text should not be longer than 50 percentage (or the passed percentage)
		if style == "title" and 50 <= (LineLength / 100 * txt_len):
			return True
		return False
	
# Gets the char count at which it shall split the text
# Returns desired split position
def split_calc_char_pos(LineLength: int, Percentage: int):
	split_pos = 100 // LineLength * Percentage
	return split_pos

def split_string_preserve_words(text, max_chars):
	"""
	Split string at provided length, preferred at whitespace.
	Returns 2 strings: line0, line1
	"""
	words = text.split()
	line0 = ""
	current_length = 0
	rem_words = []
	# Parse each word
	for word in words:
		if current_length + len(word) <= max_chars:
			line0 += word + " "
			current_length += len(word) + 1  # Add 1 for space
			rem_words.append(word)
		else:
			break
	# Remove used words
	for w in rem_words:
		words.remove(w)
	# Fill line1
	line1 = text[current_length:].strip()
	return [line0.strip(), line1.strip()]

def shorten(txt: str, char_count: int, cut_from_middle: bool = False) -> str:
	"""
	Shorten text to fit into char_count, mostly used for tui.status.
	Optional: can shorten text in the middle of the string.
	"""
	if len(txt) <= char_count:
		return txt

	if cut_from_middle:
		# TODO: Why do i need another "- 10" and a "- 12"
		# Despite the fact i had already subtracted - 12 from in tui.progress...
		remaining_chars = char_count - 10 # org - 3, then 4...
		side_chars = int(remaining_chars // 2)
		shortened_text = f"{txt[:side_chars]}...{txt[-side_chars:]}"
	else:
		# Attempt to cut at word boundaries for better readability
		words = txt.split()
		shortened_words = []
		current_length = 0
		
		for word in words:
			if current_length + len(word) <= char_count - 12:	# org 3
				shortened_words.append(word)
				current_length += len(word) + 1  # Add 1 for space between words
		
		shortened_text = ' '.join(shortened_words)
		shortened_text += '...'  # Adding ellipsis
		
	return shortened_text
################################################################################################################
#####                                           Charcount                                                  #####
################################################################################################################
def get_input_charcount(count:int) -> str:
	# Use the subprocess module to invoke the shell and read a single character
	if _IS_WINDOWS:
		chars = []
		while len(chars) < count:
			char = _msvcrt.getch()
			# Check if char is printable
			if char in _string.printable.encode():
				char_str = char.decode()
				chars.append(char_str)
				# Check for Enter key press
				if char_str == '\r' and len(chars) >= 1:
					break
		return ''.join(chars)
	else:
		# Expecting a linux based system
		result = _subprocess.run(["read", "-n", count], capture_output=True, text=True, shell=True)
		chars = result.stdout.strip()
		return chars



"""
	Created on:		2023 Nov. 09
	Created by:		Simon Arjuna Erat
	License:		MIT
	URL:			https://www.github.com/sri-arjuna/ASPIRE

	Based on my TUI & SWARM for the BASH shell © 2011
"""
################################################################################################################
#####                                            Shared Imports                                            #####
################################################################################################################
#
#	Essential imports
#
import os as _os
#
#	Cross Platform & Advanced usage
#
#import platform as _platform
#import msvcrt as _msvcrt
#import shutil as _shutil
#import subprocess as _subprocess
from typing import Union as _Union
import time as _time

#
#	Internals
#
from AspireTUI import _settings_console as _settings
from AspireTUI import _MSG
from AspireTUI.__core import _internal
from AspireTUI.__core import _PrintUtils as _put
from AspireTUI.__core import _theme as _Theme
import AspireTUI.strings as _stew

################################################################################################################
#####                                           Public Functions                                           #####
################################################################################################################
def header(*args, end='\n'):
	"""
	Expects:
		Up to 3 strings.
	Usage:
		tui.header("left", "right")
		tui.header("left", "center", "right")
		tui.header("", "", "right")
	Prints up to 3 strings, L, L+R, LCR.
	Theme:Default = White font and blue background.
	"""
	_put._update(forced=True)
	style="header"
	_put.border(style=style)
	_put.text(*args, style=style, end=end)

def title(text="", end='\n'):
	"""
	Prints 1 (one) string in the center.
	Theme:Default = Blue font and white background.
	"""
	_put._update()
	style="title"
	_put.border(style=style)
	_put.text(text, style=style, end=end)

def print(*args, end='\n'):
	"""
	Default output method.
	Prints up to 3 strings, L, L+R, LCR.
	Theme:Default = Blue font and white background for borders, default console colors for output.
	"""
	# Implementation for printe method
	_put._update()
	if len(args) == 1:
		single_arg = args[0]
		if isinstance(single_arg, str) and '\n' in single_arg:
			lines = single_arg.strip().split('\n')
			for line in lines:
				_put.border()
				_put.text(line, end='\n')
		elif _stew.isList(single_arg):
			for line in single_arg:
				_put.border()
				_put.text(line, end='\n')
		else:
			_put.border()
			_put.text(single_arg, end=end)
	else:
		_put.border()
		_put.text(*args, end=end)

def press(text=None):
	"""
	Theme conform wrapper for _os.system("pause").
	If no text is passed, "Please press any key to continue." is used.
	If passed or default text fits twice on the line, it will do so, otherwise center.
	"""
	_put._update()
	if text is None or text == "":
		text = _MSG.tui_press
	_put.border()
	if _settings["inner"] >= 2 * (len(text) + 1):
		# Fits twice
		_put.text(text, text)
	else:
		# Fits only once
		_put.text(text)
	# Workaround to hide the default message
	stdout = _os.dup(1)
	_os.dup2(_os.open(_os.devnull, _os.O_WRONLY), 1)
	_os.system("pause")
	_os.dup2(stdout, 1)

def yesno(question: str, yesno_option="yn") -> bool:
	"""
	Ask the user a simple yes/no question.
	Pass an optional 2 character-string, of which 1. is positive and 2. is negative.
	"""
	_put._update()
	style = _Theme.get()
	answer = ""
	question_string = f"{question} ({yesno_option}) {style.prompt_read} "
	yes = yesno_option[:1]
	no = yesno_option[1:]
	
	# Default Aspire / TUI output
	_put.border()
	_put.text(question_string, end="")

	# Loop for proper input:
	while True:
		# Make input ready to be checked
		answer = _internal.get_input_charcount(1) 
		#answer = input()
		if answer in yesno_option:
			break
	# Print answer
	_put._right(answer)
	# Return the according bool
	if answer in yes and "" != answer:
		return True
	elif answer in no and "" != answer:
		return False

def status(ID: _Union[int, bool], *args, align_right=True, end='\n'):
	"""
	Requires ID to be int or bool, and 1 additional string as message for the status.					\n
	By default (orientation=right), text is printed left, center, and the actual status on the right.	\n

	Change orientation=left to have status on the left, main text on the right. If an optional text is passed,
	main text will be in the center and the optional text on the right.
	"""
	if len(args) > 2:
		msg_status_many_args = _MSG.args_2_status
		raise SyntaxError(msg_status_many_args)

	_put._update()
	_put.border()

	if len(args) == 0:
		if align_right:
			_put.text("", _put.status(ID), end=end)
		else:  # left
			_put.text(_put.status(ID),  "", end=end)
	else:
		if align_right:
			_put.text(*args, _put.status(ID), end=end)
		else:  # left
			_put.text(_put.status(ID), *args, end=end)

def progress( text: str, cur: float, max: float, style: str = "bar", cut_from_end: bool = True, reverse: bool = False):
	"""
	Prints a leading 'text' with a progress indicator according to passed style.	\n
	If no 'text' is passed (read: empty-string=""), the progress bar fills all of the screen, otherwise its 50/50.
	Default style is 'bar'.		\n
	Valid styles are: bar, num	\n
	"""
	_put._update()	# DEBUG=text
	if text:
		width = _settings["inner"] / 2
	else:
		# TODO: Why do i need " - 12" for this value???
		width = _settings["inner"] - 12
	# Styles
	if "num" == style:
		prog_out = f"[ {cur} / {max} ]"
		#return True
	elif "bar" == style:
		prog_out = f"{_put.bar(cur,max,width,reverse=reverse)}"
	else:
		msg_progress_style = _MSG.tui_progress_bar
		raise ValueError(msg_progress_style)
	
	if text == "":
		text_new = ""	# There is no text anyway... idk...
	else:
		# Send to shorten anyway, 
		# this should avoid the split test in _put.text().
		if cut_from_end:
			text_new = _internal.shorten(text, width)
		else:
			text_new = _internal.shorten(text, width, True)
	_put.border()
	_put.text(text_new, prog_out, end="")
	if cur == max:
		print()
	return True

def wait(Time: _Union[float, int] ,  msg=None, unit="s",hidden=False, bar=False, iIntervall=1):
	"""
	If just used as: tui.wait(1), it waits for 1 second before it continues. \n
	To save effort, you can change unit="h" if you want to wait 1 hour instead,
	or use unit="m" to wait the passed number in minutes.

	"""
	# Init final vars:
	output = ""
	# Get measurement
	secs = Time
	org = Time
	if not unit == "s":
		secs = _stew.sec2time(secs)
		print(secs, "TODO work with other time units for wait")
		# TODO: sec2time , min2sec, hour2sec
	
	# Prepare strings:
	if msg is None:
		msg = _MSG.tui_wait_continue_in
	
	cur = secs
	while cur > 0:
		print(msg,f"{stew.sec2time(cur)} / {Time}{unit}")
		cur -= 1
		## https://realpython.com/python-sleep/
		## Async? or time.wait instead ??
		_time.sleep(iIntervall)
	

def msgbox(*args, bOutside=False):
	"""
	
	"""
	# Print "border"
	_put._update()
	if bOutside:
		header()
	else:
		title()
	# Print the content
	if len(args) > 1:
		for a in args:
			print(a)
		else:
			print(args)
	# Print "border"
	if bOutside:
		header()
	else:
		title()

def list(*args, bRoman=False, bMenu=False, sSeperator=")"):
	"""
	Prints a list 
	"""
	if bMenu:
		list_entries = [ _MSG.tui_list_back ] *args
